#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse

import wot.io


def create_parser():
    parser = argparse.ArgumentParser(description='Compute neighborhood graph')
    parser.add_argument('--matrix', help=wot.commands.MATRIX_HELP, required=True)
    parser.add_argument('--gene_filter',
                        help='File with one gene id per line to include from the matrix')
    parser.add_argument('--cell_filter',
                        help='File with one cell id per line to include from the matrix')
    parser.add_argument('--transpose', help='Transpose the matrix', action='store_true')
    parser.add_argument('--pca_comps',
                        help='Number of PCA components.',
                        type=int, default=50)
    parser.add_argument('--diff_comps',
                        help='Number of diffusion components.',
                        type=int, default=15)
    parser.add_argument('--neighbors', help='Number of nearest neighbors',
                        type=int, default=15)
    parser.add_argument('--space', help='Space to compute the neighborhood graph in', choices=['dmap', 'pca', 'input'])
    parser.add_argument('--out',
                        help='Output file name. The file is saved in gexf format (https://gephi.org/gexf/format/)')
    return parser


def main(args):
    if args.out is None:
        args.out = 'wot-neighborhood-graph'
    space = args.space
    neighbors = args.neighbors
    pca_comps = args.pca_comps
    diff_comps = args.diff_comps
    adata = wot.io.read_dataset(args.matrix)
    if args.transpose:
        adata = adata.T
    adata = wot.io.filter_adata(adata, obs_filter=args.cell_filter, var_filter=args.gene_filter)
    wot.neighborhood_graph.compute_neighborhood_graph(adata, space=space, neighbors=neighbors, pca_comps=pca_comps,
                                                      diff_comps=diff_comps)
    wot.neighborhood_graph.graph_to_gexf(adata, args.out + '.gexf')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import scanpy.api as sc


def compute_neighborhood_graph(adata, space, neighbors, pca_comps, diff_comps):
    if space == 'pca':
        sc.tl.pca(adata, n_comps=pca_comps)
        sc.pp.neighbors(adata, use_rep='X_pca', n_neighbors=neighbors)
    elif space == 'dmap':
        sc.tl.pca(adata, n_comps=pca_comps)
        sc.pp.neighbors(adata, use_rep='X_pca', n_neighbors=neighbors)
        sc.tl.diffmap(adata, n_comps=diff_comps)
        sc.pp.neighbors(adata, use_rep='X_diffmap', n_neighbors=neighbors)
    else:
        sc.pp.neighbors(adata, use_rep='X', n_neighbors=neighbors)


def graph_to_gexf(adata, path):
    W = adata.uns['neighbors']['connectivities']
    n_obs = W.shape[0]
    obs_ids = adata.obs.index.values
    with open(path, 'w') as writer:
        writer.write('<?xml version="1.0" encoding="UTF-8"?>')
        writer.write(
            '<gexf xmlns="http://www.gexf.net/1.2draft" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.gexf.net/1.2draft http://www.gexf.net/1.2draft/gexf.xsd" version="1.2">')
        writer.write('<graph defaultedgetype="undirected">')
        writer.write('<nodes>')

        for j in range(n_obs):
            writer.write('<node id="{id}" label="{id}" />'.format(id=obs_ids[j]))
        writer.write('</nodes>')
        writer.write('<edges>')
        rows, cols = W.nonzero()
        edge_counter = 1
        for i, j in zip(rows, cols):
            if i < j:
                writer.write(
                    '"<edge id="{id}" source="{s}" target="{t}" weight="{w}" />'.format(id=edge_counter, s=obs_ids[i],
                                                                                        t=obs_ids[j], w=W[i, j]))
                edge_counter = edge_counter + 1
        writer.write('</edges>')
        writer.write('</graph>')
        writer.write('</gexf>')
# <?xml version="1.0" encoding="UTF-8"?>
# <gexf xmlns="http://www.gexf.net/1.2draft" version="1.2">
#     <meta lastmodifieddate="2009-03-20">
#         <creator>Gexf.net</creator>
#         <description>A hello world! file</description>
#     </meta>
#     <graph mode="static" defaultedgetype="directed">
#         <nodes>
#             <node id="0" label="Hello" />
#             <node id="1" label="Word" />
#         </nodes>
#         <edges>
#             <edge id="0" source="0" target="1" />
#         </edges>
#     </graph>
# </gexf>



import logging

import numpy as np
import pandas as pd
import scipy.sparse
import statsmodels.stats.multitest

import wot.io

logger = logging.getLogger('wot')

import anndata


def diff_exp(adata: anndata.AnnData, fate_datasets: [anndata.AnnData], cell_days_field: str = 'day',
             nperm: int = 0, min_fold_change: float = 0, smooth_p_values: bool = False,
             compare: str = 'all', delta_days: float = None) -> {str: pd.DataFrame}:
    """
    Compute statistics for differential expression

    Parameters
    ----------
    adata :
        The expression matrix.
    fate_datasets :
        List of fate datasets.
    cell_days_field
        Cell days field in adata.
    nperm
        Number of permutations to perform
    min_fold_change
        Exclude genes for permutations that don't have at least min_fold_change
    smooth_p_values
        Whether to smooth p-values
    compare :
        within, match, all, or fate name
    delta_days
        Difference in days for comparisons

    Returns
    -------
    df :
        A dataframe with the columns name1, name2, day1, day2, distance
    """

    if cell_days_field not in adata.obs:
        raise ValueError(cell_days_field + ' not found')
    adata = anndata.AnnData(adata.X, adata.obs.copy(), adata.var)
    fate_names = []
    if not isinstance(fate_datasets, list) and not isinstance(fate_datasets, tuple):
        fate_datasets = [fate_datasets]

    # add fate dataset to obs
    for fate_ds in fate_datasets:
        fate_names += list(fate_ds.var.index)
        adata.obs = adata.obs.join(
            pd.DataFrame(index=fate_ds.obs.index, data=fate_ds.X, columns=fate_ds.var.index))

    unique_days = np.array(sorted(adata.obs[cell_days_field].unique().astype(float)))
    unique_days = unique_days[np.isnan(unique_days) == False]
    logger.info('{} days'.format(len(unique_days)))

    comparisons = wot.tmap.generate_comparisons(comparison_names=fate_names, compare=compare,
                                                days=unique_days, delta_days=delta_days, reference_day='start')

    current_name1 = None
    current_name2 = None
    df = None
    features = adata.var.index
    results = {}
    for comparison in comparisons:
        names = comparison[0]
        days = comparison[1]
        name1 = names[0]
        name2 = names[1]
        day1 = days[0]
        day2 = days[1]
        if current_name1 != name1 or current_name2 != name2:
            if df is not None:
                results['{}_{}'.format(current_name1, current_name2).replace('/', '-')] = df

            df = pd.DataFrame(index=features)
            current_name1 = name1
            current_name2 = name2

        logger.info('{} vs {}, day {}, day {}'.format(name1, name2, day1, day2))
        values1, weights1 = __get_expression_and_weights(adata, cell_days_field, day1, name1)
        values2, weights2 = __get_expression_and_weights(adata, cell_days_field, day2, name2)

        df = __add_stats(values1, weights1, df, '_{}_{}'.format(name1, day1), features)

        df = __add_stats(values2, weights2, df, '_{}_{}'.format(name2, day2), features)

        df = df.join(__do_comparison(values1, weights1, day1, values2, weights2, day2, nperm=nperm,
                                     min_fold_change=min_fold_change,
                                     smooth_p_values=smooth_p_values, features=features))

    if df is not None:
        results['{}_{}'.format(current_name1, current_name2).replace('/', '-')] = df
    return results


def __add_stats(expression_values, weights, df, suffix, features):
    # expression_values = np.expm1(expression_values)
    mean = np.average(expression_values, weights=weights, axis=0)
    fraction_expressed = weights.dot(expression_values > 0)
    # variance = np.average((expression_values - mean) ** 2, weights=weights, axis=0)
    # variance = np.log1p(variance)
    # mean = np.log1p(mean)
    if 'mean{}'.format(suffix) not in df:
        return df.join(pd.DataFrame(index=features,
                                    data={
                                        'mean{}'.format(suffix): mean,
                                        'fraction_expressed{}'.format(suffix): fraction_expressed
                                    }))
    return df


def __get_expression_and_weights(adata, cell_days_field, day, fate_name):
    ds = adata[
        (adata.obs[cell_days_field] == day) & (
                False == adata.obs[fate_name].isna())]
    weights = ds.obs[fate_name].values
    expression_values = ds.X
    if scipy.sparse.isspmatrix(expression_values):
        expression_values = expression_values.toarray()
    weights = weights / weights.sum()
    return expression_values, weights


def __do_comparison(expression_values1, weights1, day1, expression_values2, weights2, day2, nperm, min_fold_change,
                    smooth_p_values, features, fraction_expressed_ratio_add=0.0001):
    # expression_values1 = np.expm1(expression_values1)
    # expression_values2 = np.expm1(expression_values2)
    mean1 = np.average(expression_values1, weights=weights1, axis=0)
    mean2 = np.average(expression_values2, weights=weights2, axis=0)
    fraction_expressed1 = weights1.dot(expression_values1 > 0)
    fraction_expressed2 = weights2.dot(expression_values2 > 0)
    fraction_expressed_diff = (fraction_expressed1 + fraction_expressed_ratio_add) / (
            fraction_expressed2 + fraction_expressed_ratio_add)

    # fold_change = np.log1p(mean1) - np.log1p(mean2)
    observed = (mean1 - mean2)
    suffix = "_{}_{}".format(day1, day2)
    results = pd.DataFrame(index=features,
                           data={'fold_change' + suffix: observed,
                                 'fraction_expressed_ratio' + suffix: fraction_expressed_diff})

    if nperm is not None and nperm > 0:
        genes_use = np.abs(observed) >= min_fold_change
        if genes_use.sum() > 0:
            expression_values1 = expression_values1[:, genes_use]
            expression_values2 = expression_values2[:, genes_use]
            observed_use = observed[genes_use]
            weights1 = weights1.copy()
            weights2 = weights2.copy()
            p = np.zeros(expression_values2.shape[1])

            for i in range(nperm):
                np.random.shuffle(weights1)
                np.random.shuffle(weights2)
                mean1 = np.average(expression_values1, weights=weights1, axis=0)
                mean2 = np.average(expression_values2, weights=weights2, axis=0)
                permuted = (mean1 - mean2)
                p[(permuted >= observed_use)] += 1
            # 2-sided p-value
            k = p
            if smooth_p_values:
                p = (p + 1) / (nperm + 2)
            else:
                p = p / nperm
            one_minus_p = 1.0 - p;
            expr = one_minus_p < p
            p[expr] = 1 - p[expr]
            p *= 2
            fdr = statsmodels.stats.multitest.multipletests(p)[1]
            results = results.join(
                pd.DataFrame(index=features[genes_use], data={'p_value' + suffix: p,
                                                              'fdr' + suffix: fdr,
                                                              'k' + suffix: k}))
    return results



import anndata

""" Reads in a gct file .

The main method is parse. parse_into_3 creates the row
metadata, column metadata, and data dataframes, while the
assemble_multi_index_df method in GCToo.py assembles them.

1) Example GCT v1.3:
        ----- start of file ------
        #1.3
        96 36 9 15
        ---------------------------------------------------
        |id|        rhd          |          cid           |
        ---------------------------------------------------
        |  |                     |                        |
        |c |                     |                        |
        |h |      (blank)        |      col_metadata      |
        |d |                     |                        |
        |  |                     |                        |
        ---------------------------------------------------
        |  |                     |                        |
        |r |                     |                        |
        |i |    row_metadata     |          data          |
        |d |                     |                        |
        |  |                     |                        |
        ---------------------------------------------------
        ----- end of file ------

        Notes:
        - line 1 of file ("#1.3") refers to the version number
        - line 2 of file ("96 36 9 15") refers to the following:
                -96 = number of data rows
                -36 = number of data columns
                -9 = number of row metadata fields (+1 for the 'id' column -- first column)
                -15 = number of col metadata fields (+1 for the 'id' row -- first row)
        - Once read into a DataFrame, col_metadata_df is stored as the transpose of how it looks in the gct file.
                That is, col_metadata_df.shape = (num_cid, num_chd).

2) Example GCT v1.2

        ----- start of file ------
        #1.2
        96 36
        -----------------------------------------------
        |"NAME" |"Description"|          cid           |
        -----------------------------------------------
        |   r   |             |                        |
        |   i   |             |                        |
        |   d   |row_metadata |         data           |
        |       |             |                        |
        |       |             |                        |
        -----------------------------------------------

        ----- end of file ------
        Notes:
        - line 1 of file ("#1.3") refers to the version number
        - line 2 of file ("96 36 9 15") refers to the following:
                -96 = number of data rows
                -36 = number of data columns

"""

import numpy as np
import pandas as pd

__author__ = "Lev Litichevskiy, Oana Enache"
__email__ = "lev@broadinstitute.org"


def read_gct(file_path):
    """ The main method.

    Args:
        - file_path (string): full path to gct file you want to parse

    Returns:
        anndata.AnnData

    """

    # Read version and dimensions
    (version, num_data_rows, num_data_cols,
     num_row_metadata, num_col_metadata) = read_version_and_dims(file_path)

    # Read in metadata and data
    (row_metadata, col_metadata, data) = parse_into_3(
        file_path, num_data_rows, num_data_cols,
        num_row_metadata, num_col_metadata)

    row_metadata.index.name = None
    col_metadata.index.name = None
    row_metadata.columns.name = None
    col_metadata.columns.name = None
    return anndata.AnnData(data, row_metadata, col_metadata)


def read_version_and_dims(file_path):
    # Open file
    f = open(file_path, "r")

    # Get version from the first line
    version = f.readline().strip().lstrip("#")

    if version not in ["1.3", "1.2"]:
        err_msg = ("Only GCT1.2 and 1.3 are supported. The first row of the GCT " +
                   "file must simply be (without quotes) '#1.3' or '#1.2'")

        raise Exception(err_msg.format(version))

    # Convert version to a string
    version_as_string = "GCT" + str(version)

    # Read dimensions from the second line
    dims = f.readline().strip().split("\t")

    # Close file
    f.close()

    # Check that the second row is what we expect
    if version == "1.2" and len(dims) != 2:
        error_msg = "GCT1.2 should have 2 dimension-related entries in row 2. dims: {}"

        raise Exception(error_msg.format(dims))
    elif version == "1.3" and len(dims) != 4:
        error_msg = "GCT1.3 should have 4 dimension-related entries in row 2. dims: {}"

        raise Exception(error_msg.format(dims))

    # Explicitly define each dimension
    num_data_rows = int(dims[0])
    num_data_cols = int(dims[1])
    if len(dims) == 4:
        num_row_metadata = int(dims[2])
        num_col_metadata = int(dims[3])
    else:
        num_row_metadata = 1
        num_col_metadata = 0

        # Return version and dimensions
    return version_as_string, num_data_rows, num_data_cols, num_row_metadata, num_col_metadata


def parse_into_3(file_path, num_data_rows, num_data_cols, num_row_metadata, num_col_metadata):
    # Read the gct file beginning with line 3
    full_df = pd.read_csv(file_path, sep="\t", header=None, skiprows=2, dtype=str)

    # Check that full_df is the size we expect
    assert full_df.shape == (num_col_metadata + num_data_rows + 1,
                             num_row_metadata + num_data_cols + 1), (
        ("The shape of full_df is not as expected: data is {} x {} " +
         "but there are {} row meta fields and {} col fields").format(
            num_data_rows, num_data_cols, num_row_metadata, num_col_metadata))

    # Assemble metadata dataframes
    row_metadata = assemble_row_metadata(full_df, num_col_metadata, num_data_rows, num_row_metadata)
    col_metadata = assemble_col_metadata(full_df, num_col_metadata, num_row_metadata, num_data_cols)

    # Assemble data dataframe
    data = assemble_data(full_df, num_col_metadata, num_data_rows, num_row_metadata, num_data_cols)

    # Return 3 dataframes
    return row_metadata, col_metadata, data


def assemble_row_metadata(full_df, num_col_metadata, num_data_rows, num_row_metadata):
    # Extract values
    row_metadata_row_inds = range(num_col_metadata + 1, num_col_metadata + num_data_rows + 1)
    row_metadata_col_inds = range(1, num_row_metadata + 1)
    row_metadata = full_df.iloc[row_metadata_row_inds, row_metadata_col_inds]

    # Create index from the first column of full_df (after the filler block)
    row_metadata.set_index(full_df.iloc[row_metadata_row_inds, 0], inplace=True)
    # Create columns from the top row of full_df (before cids start)
    row_metadata.columns = full_df.iloc[0, row_metadata_col_inds].values

    # Rename the index name and columns name
    # row_metadata.index.name = row_index_name
    # row_metadata.columns.name = row_header_name

    # Convert metadata to numeric if possible
    row_metadata = row_metadata.apply(lambda x: pd.to_numeric(x, errors="ignore"))

    return row_metadata


def assemble_col_metadata(full_df, num_col_metadata, num_row_metadata, num_data_cols):
    # Extract values
    col_metadata_row_inds = range(1, num_col_metadata + 1)
    col_metadata_col_inds = range(num_row_metadata + 1, num_row_metadata + num_data_cols + 1)
    col_metadata = full_df.iloc[col_metadata_row_inds, col_metadata_col_inds]

    # Transpose so that samples are the rows and headers are the columns
    col_metadata = col_metadata.T

    # Create index from the top row of full_df (after the filler block)
    col_metadata.set_index(full_df.iloc[0, col_metadata_col_inds], inplace=True)

    # Create columns from the first column of full_df (before rids start)
    col_metadata.columns = full_df.iloc[col_metadata_row_inds, 0].values

    # # Rename the index name and columns name
    # col_metadata.index.name = 'id'
    # col_metadata.columns.name = 'description'

    # Convert metadata to numeric if possible
    col_metadata = col_metadata.apply(lambda x: pd.to_numeric(x, errors="ignore"))

    return col_metadata


def assemble_data(full_df, num_col_metadata, num_data_rows, num_row_metadata, num_data_cols):
    # Extract values
    data_row_inds = range(num_col_metadata + 1, num_col_metadata + num_data_rows + 1)
    data_col_inds = range(num_row_metadata + 1, num_row_metadata + num_data_cols + 1)
    data = full_df.iloc[data_row_inds, data_col_inds].values
    return data.astype(np.float64)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = [
    'numpy', 'pandas', 'h5py', 'scanpy>=1.4.2', 'scikit-learn', 'scipy', 'matplotlib', 'POT'
]

extras_require = {
}

setup_requirements = [
    'Cython'
]

test_requirements = [
    'unittest'
]

setuptools.setup(
    name='wot',
    version='1.0.3',
    description="Optimal transport for time-course single cell data",
    author="WOT Team",
    author_email='wot@broadinstitute.org',
    url='https://github.com/broadinstitute/wot',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(include=['wot']),
    include_package_data=True,
    install_requires=requirements,
    license="BSD license",
    zip_safe=False,
    keywords='wot',
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Visualization',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
    python_requires='>= 3',
    entry_points={
        'console_scripts': [
            'wot=wot.__main__:main'
        ]
    },
)



# -*- coding: utf-8 -*-
import wot.graphics
import wot.io
import wot.ot
import wot.simulate
import wot.tmap
from .dataset_util import *
from .gene_set_scores import *
from .neighborhood_graph import *
from .population import *
